"""Convenience functions for creating MCEs"""

import time
from typing import List, Union

from datahub.metadata import (
    AuditStampClass,
    DatasetLineageTypeClass,
    DatasetSnapshotClass,
    MetadataChangeEventClass,
    UpstreamClass,
    UpstreamLineageClass,
)

DEFAULT_ENV = "PROD"
DEFAULT_FLOW_CLUSTER = "prod"


def make_dataset_urn(platform: str, name: str, env: str = DEFAULT_ENV):
    return f"urn:li:dataset:(urn:li:dataPlatform:{platform},{name},{env})"


def make_user_urn(username: str):
    return f"urn:li:corpuser:{username}"


def make_data_flow_urn(
    orchestrator: str, flow_id: str, cluster: str = DEFAULT_FLOW_CLUSTER
):
    return f"urn:li:dataFlow:({orchestrator},{flow_id},{cluster})"


def make_data_job_urn_with_flow(flow_urn: str, job_id: str):
    return f"urn:li:dataJob:({flow_urn},{job_id})"


def make_data_job_urn(
    orchestrator: str, flow_id: str, job_id: str, cluster: str = DEFAULT_FLOW_CLUSTER
):
    return make_data_job_urn_with_flow(
        make_data_flow_urn(orchestrator, flow_id, cluster), job_id
    )


def make_lineage_mce(
    upstream_urns: Union[str, List[str]],
    downstream_urn: str,
    actor: str = make_user_urn("datahub"),
    lineage_type: str = DatasetLineageTypeClass.TRANSFORMED,
) -> MetadataChangeEventClass:
    sys_time = int(time.time() * 1000)
    if not isinstance(upstream_urns, list):
        upstream_urns = [upstream_urns]

    mce = MetadataChangeEventClass(
        proposedSnapshot=DatasetSnapshotClass(
            urn=downstream_urn,
            aspects=[
                UpstreamLineageClass(
                    upstreams=[
                        UpstreamClass(
                            auditStamp=AuditStampClass(
                                time=sys_time,
                                actor=actor,
                            ),
                            dataset=upstream_urn,
                            type=lineage_type,
                        )
                        for upstream_urn in upstream_urns
                    ]
                )
            ],
        )
    )
    return mce



from dataclasses import dataclass
from typing import Iterable, Optional

import ldap
from ldap.controls import SimplePagedResultsControl

from datahub.configuration.common import ConfigModel, ConfigurationError
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.source.metadata_common import MetadataWorkUnit
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.schema_classes import CorpUserInfoClass, CorpUserSnapshotClass


def create_controls(pagesize):
    """
    Create an LDAP control with a page size of "pagesize".
    """
    return SimplePagedResultsControl(True, size=pagesize, cookie="")


def get_pctrls(serverctrls):
    """
    Lookup an LDAP paged control object from the returned controls.
    """
    return [
        c for c in serverctrls if c.controlType == SimplePagedResultsControl.controlType
    ]


def set_cookie(lc_object, pctrls, pagesize):
    """
    Push latest cookie back into the page control.
    """
    cookie = pctrls[0].cookie
    lc_object.cookie = cookie
    return cookie


def guess_person_ldap(dn, attrs) -> Optional[str]:
    """Determine the user's LDAP based on the DN and attributes."""
    if "sAMAccountName" in attrs:
        return attrs["sAMAccountName"][0].decode()
    if "uid" in attrs:
        return attrs["uid"][0].decode()
    return None


class LDAPSourceConfig(ConfigModel):
    # Server configuration.
    ldap_server: str
    ldap_user: str
    ldap_password: str

    # Extraction configuration.
    base_dn: str
    filter: str = "(objectClass=*)"

    page_size: int = 20


@dataclass
class LDAPSource(Source):
    config: LDAPSourceConfig
    report: SourceReport

    def __init__(self, ctx: PipelineContext, config: LDAPSourceConfig):
        super().__init__(ctx)
        self.config = config
        self.report = SourceReport()

        ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_ALLOW)
        ldap.set_option(ldap.OPT_REFERRALS, 0)

        self.ldap_client = ldap.initialize(self.config.ldap_server)
        self.ldap_client.protocol_version = 3

        try:
            self.ldap_client.simple_bind_s(
                self.config.ldap_user, self.config.ldap_password
            )
        except ldap.LDAPError as e:
            raise ConfigurationError("LDAP connection failed") from e

        self.lc = create_controls(self.config.page_size)

    @classmethod
    def create(cls, config_dict, ctx):
        config = LDAPSourceConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_workunits(self) -> Iterable[MetadataWorkUnit]:
        cookie = True
        while cookie:
            try:
                msgid = self.ldap_client.search_ext(
                    self.config.base_dn,
                    ldap.SCOPE_SUBTREE,
                    self.config.filter,
                    serverctrls=[self.lc],
                )
                rtype, rdata, rmsgid, serverctrls = self.ldap_client.result3(msgid)
            except ldap.LDAPError as e:
                self.report.report_failure(
                    "ldap-control", "LDAP search failed: {}".format(e)
                )
                break

            for dn, attrs in rdata:
                # TODO: create groups if 'organizationalUnit' in attrs['objectClass']

                if (
                    b"inetOrgPerson" in attrs["objectClass"]
                    or b"posixAccount" in attrs["objectClass"]
                ):
                    yield from self.handle_user(dn, attrs)

            pctrls = get_pctrls(serverctrls)
            if not pctrls:
                self.report.report_failure(
                    "ldap-control", "Server ignores RFC 2696 control."
                )
                break

            cookie = set_cookie(self.lc, pctrls, self.config.page_size)

    def handle_user(self, dn, attrs) -> Iterable[MetadataWorkUnit]:
        """
        Handle a DN and attributes by adding manager info and constructing a
        work unit based on the information.
        """
        manager_ldap = None
        if "manager" in attrs:
            try:
                m_cn = attrs["manager"][0].split(b",")[0]
                manager_msgid = self.ldap_client.search_ext(
                    self.config.base_dn,
                    ldap.SCOPE_SUBTREE,
                    f"({m_cn.decode()})",
                    serverctrls=[self.lc],
                )
                m_dn, m_attrs = self.ldap_client.result3(manager_msgid)[1][0]
                manager_ldap = guess_person_ldap(m_dn, m_attrs)
            except ldap.LDAPError as e:
                self.report.report_warning(
                    dn, "manager LDAP search failed: {}".format(e)
                )

        mce = self.build_corp_user_mce(dn, attrs, manager_ldap)
        if mce:
            wu = MetadataWorkUnit(dn, mce)
            self.report.report_workunit(wu)
            yield wu
        yield from []

    def build_corp_user_mce(
        self, dn, attrs, manager_ldap
    ) -> Optional[MetadataChangeEvent]:
        """
        Create the MetadataChangeEvent via DN and attributes.
        """
        ldap = guess_person_ldap(dn, attrs)
        full_name = attrs["cn"][0].decode()
        first_name = attrs["givenName"][0].decode()
        last_name = attrs["sn"][0].decode()
        email = (attrs["mail"][0]).decode() if "mail" in attrs else None
        display_name = (
            (attrs["displayName"][0]).decode() if "displayName" in attrs else full_name
        )
        department = (
            (attrs["departmentNumber"][0]).decode()
            if "departmentNumber" in attrs
            else None
        )
        title = attrs["title"][0].decode() if "title" in attrs else None
        manager_urn = f"urn:li:corpuser:{manager_ldap}" if manager_ldap else None

        mce = MetadataChangeEvent(
            proposedSnapshot=CorpUserSnapshotClass(
                urn=f"urn:li:corpuser:{ldap}",
                aspects=[
                    CorpUserInfoClass(
                        active=True,
                        email=email,
                        fullName=full_name,
                        firstName=first_name,
                        lastName=last_name,
                        departmentName=department,
                        displayName=display_name,
                        title=title,
                        managerUrn=manager_urn,
                    )
                ],
            )
        )

        return mce

    def get_report(self):
        return self.report

    def close(self):
        self.ldap_client.unbind()



from typing import Any, Dict, List, Union

from airflow.exceptions import AirflowException

try:
    from airflow.hooks.base import BaseHook

    AIRFLOW_1 = False
except ImportError:
    from airflow.hooks.base_hook import BaseHook

    AIRFLOW_1 = True

from datahub.emitter.kafka_emitter import DatahubKafkaEmitter
from datahub.emitter.rest_emitter import DatahubRestEmitter
from datahub.ingestion.sink.datahub_kafka import KafkaSinkConfig
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent

_default_hook_args = []
if AIRFLOW_1:
    _default_hook_args = [None]


class DatahubRestHook(BaseHook):
    conn_name_attr = "datahub_rest_conn_id"
    default_conn_name = "datahub_rest_default"
    conn_type = "datahub_rest"
    hook_name = "DataHub REST Server"

    def __init__(self, datahub_rest_conn_id: str = default_conn_name) -> None:
        super().__init__(*_default_hook_args)
        self.datahub_rest_conn_id = datahub_rest_conn_id

    @staticmethod
    def get_connection_form_widgets() -> Dict[str, Any]:
        return {}

    @staticmethod
    def get_ui_field_behaviour() -> Dict:
        """Returns custom field behavior"""
        return {
            "hidden_fields": ["port", "schema", "login"],
            "relabeling": {
                "host": "Server Endpoint",
            },
        }

    def _gms_endpoint(self) -> str:
        conn = self.get_connection(self.datahub_rest_conn_id)
        host = conn.host
        if host is None:
            raise AirflowException("host parameter is required")
        return host

    def make_emitter(self) -> DatahubRestEmitter:
        return DatahubRestEmitter(self._gms_endpoint())

    def emit_mces(self, mces: List[MetadataChangeEvent]) -> None:
        emitter = self.make_emitter()

        for mce in mces:
            emitter.emit_mce(mce)


class DatahubKafkaHook(BaseHook):
    conn_name_attr = "datahub_kafka_conn_id"
    default_conn_name = "datahub_kafka_default"
    conn_type = "datahub_kafka"
    hook_name = "DataHub Kafka Sink"

    def __init__(self, datahub_kafka_conn_id: str = default_conn_name) -> None:
        super().__init__(*_default_hook_args)
        self.datahub_kafka_conn_id = datahub_kafka_conn_id

    @staticmethod
    def get_connection_form_widgets() -> Dict[str, Any]:
        return {}

    @staticmethod
    def get_ui_field_behaviour() -> Dict:
        """Returns custom field behavior"""
        return {
            "hidden_fields": ["port", "schema", "login", "password"],
            "relabeling": {
                "host": "Kafka Broker",
            },
        }

    def _get_config(self) -> KafkaSinkConfig:
        conn = self.get_connection(self.datahub_kafka_conn_id)
        obj = conn.extra_dejson
        obj.setdefault("connection", {})
        if conn.host is not None:
            if "bootstrap" in obj["connection"]:
                raise AirflowException(
                    "Kafka broker specified twice (present in host and extra)"
                )
            obj["connection"]["bootstrap"] = conn.host
        config = KafkaSinkConfig.parse_obj(obj)
        return config

    def make_emitter(self) -> DatahubKafkaEmitter:
        sink_config = self._get_config()
        return DatahubKafkaEmitter(sink_config)

    def emit_mces(self, mces: List[MetadataChangeEvent]) -> None:
        emitter = self.make_emitter()
        errors = []

        def callback(exc, msg):
            if exc:
                errors.append(exc)

        for mce in mces:
            emitter.emit_mce_async(mce, callback)

        emitter.flush()

        if errors:
            raise AirflowException(f"failed to push some MCEs: {errors}")


class DatahubGenericHook(BaseHook):
    def __init__(self, datahub_conn_id: str) -> None:
        super().__init__(*_default_hook_args)
        self.datahub_conn_id = datahub_conn_id

    def get_underlying_hook(self) -> Union[DatahubRestHook, DatahubKafkaHook]:
        conn = self.get_connection(self.datahub_conn_id)
        if conn.conn_type == DatahubRestHook.conn_type:
            return DatahubRestHook(self.datahub_conn_id)
        elif conn.conn_type == DatahubKafkaHook.conn_type:
            return DatahubKafkaHook(self.datahub_conn_id)
        else:
            raise AirflowException(
                f"DataHub cannot handle conn_type {conn.conn_type} in {conn}"
            )

    def make_emitter(self) -> Union[DatahubRestEmitter, DatahubKafkaEmitter]:
        return self.get_underlying_hook().make_emitter()

    def emit_mces(self, mces: List[MetadataChangeEvent]) -> None:
        return self.get_underlying_hook().emit_mces(mces)



# This import verifies that the dependencies are available.
import psycopg2  # noqa: F401

# GeoAlchemy adds support for PostGIS extensions in SQLAlchemy. In order to
# activate it, we must import it so that it can hook into SQLAlchemy. While
# we don't use the Geometry type that we import, we do care about the side
# effects of the import. For more details, see here:
# https://geoalchemy-2.readthedocs.io/en/latest/core_tutorial.html#reflecting-tables.
from geoalchemy2 import Geometry  # noqa: F401

from .sql_common import BasicSQLAlchemyConfig, SQLAlchemySource


class PostgresConfig(BasicSQLAlchemyConfig):
    # defaults
    scheme = "postgresql+psycopg2"


class PostgresSource(SQLAlchemySource):
    def __init__(self, config, ctx):
        super().__init__(config, ctx, "postgresql")

    @classmethod
    def create(cls, config_dict, ctx):
        config = PostgresConfig.parse_obj(config_dict)
        return cls(config, ctx)



# flake8: noqa
# fmt: off
from .schema_classes import SCHEMA as get_schema_type
from .schema_classes import _json_converter as json_converter
from .schema_classes import KafkaAuditHeaderClass
from .schema_classes import ChartInfoClass
from .schema_classes import ChartQueryClass
from .schema_classes import ChartQueryTypeClass
from .schema_classes import ChartTypeClass
from .schema_classes import AccessLevelClass
from .schema_classes import AuditStampClass
from .schema_classes import ChangeAuditStampsClass
from .schema_classes import CostClass
from .schema_classes import CostCostClass
from .schema_classes import CostCostDiscriminatorClass
from .schema_classes import CostTypeClass
from .schema_classes import DeprecationClass
from .schema_classes import GlobalTagsClass
from .schema_classes import InstitutionalMemoryClass
from .schema_classes import InstitutionalMemoryMetadataClass
from .schema_classes import MLFeatureDataTypeClass
from .schema_classes import OwnerClass
from .schema_classes import OwnershipClass
from .schema_classes import OwnershipSourceClass
from .schema_classes import OwnershipSourceTypeClass
from .schema_classes import OwnershipTypeClass
from .schema_classes import StatusClass
from .schema_classes import TagAssociationClass
from .schema_classes import VersionTagClass
from .schema_classes import TransformationTypeClass
from .schema_classes import UDFTransformerClass
from .schema_classes import DashboardInfoClass
from .schema_classes import DataFlowInfoClass
from .schema_classes import DataJobInfoClass
from .schema_classes import DataJobInputOutputClass
from .schema_classes import AzkabanJobTypeClass
from .schema_classes import DataProcessInfoClass
from .schema_classes import DatasetDeprecationClass
from .schema_classes import DatasetFieldMappingClass
from .schema_classes import DatasetLineageTypeClass
from .schema_classes import DatasetPropertiesClass
from .schema_classes import DatasetUpstreamLineageClass
from .schema_classes import UpstreamClass
from .schema_classes import UpstreamLineageClass
from .schema_classes import CorpGroupInfoClass
from .schema_classes import CorpUserEditableInfoClass
from .schema_classes import CorpUserInfoClass
from .schema_classes import ChartSnapshotClass
from .schema_classes import CorpGroupSnapshotClass
from .schema_classes import CorpUserSnapshotClass
from .schema_classes import DashboardSnapshotClass
from .schema_classes import DataFlowSnapshotClass
from .schema_classes import DataJobSnapshotClass
from .schema_classes import DataProcessSnapshotClass
from .schema_classes import DatasetSnapshotClass
from .schema_classes import MLFeatureSnapshotClass
from .schema_classes import MLModelSnapshotClass
from .schema_classes import TagSnapshotClass
from .schema_classes import BaseDataClass
from .schema_classes import CaveatDetailsClass
from .schema_classes import CaveatsAndRecommendationsClass
from .schema_classes import EthicalConsiderationsClass
from .schema_classes import EvaluationDataClass
from .schema_classes import IntendedUseClass
from .schema_classes import IntendedUserTypeClass
from .schema_classes import MLFeaturePropertiesClass
from .schema_classes import MLModelFactorPromptsClass
from .schema_classes import MLModelFactorsClass
from .schema_classes import MLModelPropertiesClass
from .schema_classes import MetricsClass
from .schema_classes import QuantitativeAnalysesClass
from .schema_classes import SourceCodeClass
from .schema_classes import SourceCodeUrlClass
from .schema_classes import SourceCodeUrlTypeClass
from .schema_classes import TrainingDataClass
from .schema_classes import MetadataChangeEventClass
from .schema_classes import ArrayTypeClass
from .schema_classes import BinaryJsonSchemaClass
from .schema_classes import BooleanTypeClass
from .schema_classes import BytesTypeClass
from .schema_classes import DatasetFieldForeignKeyClass
from .schema_classes import DateTypeClass
from .schema_classes import EditableSchemaFieldInfoClass
from .schema_classes import EditableSchemaMetadataClass
from .schema_classes import EnumTypeClass
from .schema_classes import EspressoSchemaClass
from .schema_classes import FixedTypeClass
from .schema_classes import ForeignKeySpecClass
from .schema_classes import KafkaSchemaClass
from .schema_classes import KeyValueSchemaClass
from .schema_classes import MapTypeClass
from .schema_classes import MySqlDDLClass
from .schema_classes import NullTypeClass
from .schema_classes import NumberTypeClass
from .schema_classes import OracleDDLClass
from .schema_classes import OrcSchemaClass
from .schema_classes import OtherSchemaClass
from .schema_classes import PrestoDDLClass
from .schema_classes import RecordTypeClass
from .schema_classes import SchemaFieldClass
from .schema_classes import SchemaFieldDataTypeClass
from .schema_classes import SchemaMetadataClass
from .schema_classes import SchemalessClass
from .schema_classes import StringTypeClass
from .schema_classes import TimeTypeClass
from .schema_classes import UnionTypeClass
from .schema_classes import UrnForeignKeyClass
from .schema_classes import TagPropertiesClass
from avro.io import DatumReader


class SpecificDatumReader(DatumReader):
    SCHEMA_TYPES = {
        "KafkaAuditHeader": KafkaAuditHeaderClass,
        ".KafkaAuditHeader": KafkaAuditHeaderClass,
        "com.linkedin.events.KafkaAuditHeader": KafkaAuditHeaderClass,
        "ChartInfo": ChartInfoClass,
        ".ChartInfo": ChartInfoClass,
        "com.linkedin.pegasus2avro.chart.ChartInfo": ChartInfoClass,
        "ChartQuery": ChartQueryClass,
        ".ChartQuery": ChartQueryClass,
        "com.linkedin.pegasus2avro.chart.ChartQuery": ChartQueryClass,
        "ChartQueryType": ChartQueryTypeClass,
        ".ChartQueryType": ChartQueryTypeClass,
        "com.linkedin.pegasus2avro.chart.ChartQueryType": ChartQueryTypeClass,
        "ChartType": ChartTypeClass,
        ".ChartType": ChartTypeClass,
        "com.linkedin.pegasus2avro.chart.ChartType": ChartTypeClass,
        "AccessLevel": AccessLevelClass,
        ".AccessLevel": AccessLevelClass,
        "com.linkedin.pegasus2avro.common.AccessLevel": AccessLevelClass,
        "AuditStamp": AuditStampClass,
        ".AuditStamp": AuditStampClass,
        "com.linkedin.pegasus2avro.common.AuditStamp": AuditStampClass,
        "ChangeAuditStamps": ChangeAuditStampsClass,
        ".ChangeAuditStamps": ChangeAuditStampsClass,
        "com.linkedin.pegasus2avro.common.ChangeAuditStamps": ChangeAuditStampsClass,
        "Cost": CostClass,
        ".Cost": CostClass,
        "com.linkedin.pegasus2avro.common.Cost": CostClass,
        "CostCost": CostCostClass,
        ".CostCost": CostCostClass,
        "com.linkedin.pegasus2avro.common.CostCost": CostCostClass,
        "CostCostDiscriminator": CostCostDiscriminatorClass,
        ".CostCostDiscriminator": CostCostDiscriminatorClass,
        "com.linkedin.pegasus2avro.common.CostCostDiscriminator": CostCostDiscriminatorClass,
        "CostType": CostTypeClass,
        ".CostType": CostTypeClass,
        "com.linkedin.pegasus2avro.common.CostType": CostTypeClass,
        "Deprecation": DeprecationClass,
        ".Deprecation": DeprecationClass,
        "com.linkedin.pegasus2avro.common.Deprecation": DeprecationClass,
        "GlobalTags": GlobalTagsClass,
        ".GlobalTags": GlobalTagsClass,
        "com.linkedin.pegasus2avro.common.GlobalTags": GlobalTagsClass,
        "InstitutionalMemory": InstitutionalMemoryClass,
        ".InstitutionalMemory": InstitutionalMemoryClass,
        "com.linkedin.pegasus2avro.common.InstitutionalMemory": InstitutionalMemoryClass,
        "InstitutionalMemoryMetadata": InstitutionalMemoryMetadataClass,
        ".InstitutionalMemoryMetadata": InstitutionalMemoryMetadataClass,
        "com.linkedin.pegasus2avro.common.InstitutionalMemoryMetadata": InstitutionalMemoryMetadataClass,
        "MLFeatureDataType": MLFeatureDataTypeClass,
        ".MLFeatureDataType": MLFeatureDataTypeClass,
        "com.linkedin.pegasus2avro.common.MLFeatureDataType": MLFeatureDataTypeClass,
        "Owner": OwnerClass,
        ".Owner": OwnerClass,
        "com.linkedin.pegasus2avro.common.Owner": OwnerClass,
        "Ownership": OwnershipClass,
        ".Ownership": OwnershipClass,
        "com.linkedin.pegasus2avro.common.Ownership": OwnershipClass,
        "OwnershipSource": OwnershipSourceClass,
        ".OwnershipSource": OwnershipSourceClass,
        "com.linkedin.pegasus2avro.common.OwnershipSource": OwnershipSourceClass,
        "OwnershipSourceType": OwnershipSourceTypeClass,
        ".OwnershipSourceType": OwnershipSourceTypeClass,
        "com.linkedin.pegasus2avro.common.OwnershipSourceType": OwnershipSourceTypeClass,
        "OwnershipType": OwnershipTypeClass,
        ".OwnershipType": OwnershipTypeClass,
        "com.linkedin.pegasus2avro.common.OwnershipType": OwnershipTypeClass,
        "Status": StatusClass,
        ".Status": StatusClass,
        "com.linkedin.pegasus2avro.common.Status": StatusClass,
        "TagAssociation": TagAssociationClass,
        ".TagAssociation": TagAssociationClass,
        "com.linkedin.pegasus2avro.common.TagAssociation": TagAssociationClass,
        "VersionTag": VersionTagClass,
        ".VersionTag": VersionTagClass,
        "com.linkedin.pegasus2avro.common.VersionTag": VersionTagClass,
        "TransformationType": TransformationTypeClass,
        ".TransformationType": TransformationTypeClass,
        "com.linkedin.pegasus2avro.common.fieldtransformer.TransformationType": TransformationTypeClass,
        "UDFTransformer": UDFTransformerClass,
        ".UDFTransformer": UDFTransformerClass,
        "com.linkedin.pegasus2avro.common.fieldtransformer.UDFTransformer": UDFTransformerClass,
        "DashboardInfo": DashboardInfoClass,
        ".DashboardInfo": DashboardInfoClass,
        "com.linkedin.pegasus2avro.dashboard.DashboardInfo": DashboardInfoClass,
        "DataFlowInfo": DataFlowInfoClass,
        ".DataFlowInfo": DataFlowInfoClass,
        "com.linkedin.pegasus2avro.datajob.DataFlowInfo": DataFlowInfoClass,
        "DataJobInfo": DataJobInfoClass,
        ".DataJobInfo": DataJobInfoClass,
        "com.linkedin.pegasus2avro.datajob.DataJobInfo": DataJobInfoClass,
        "DataJobInputOutput": DataJobInputOutputClass,
        ".DataJobInputOutput": DataJobInputOutputClass,
        "com.linkedin.pegasus2avro.datajob.DataJobInputOutput": DataJobInputOutputClass,
        "AzkabanJobType": AzkabanJobTypeClass,
        ".AzkabanJobType": AzkabanJobTypeClass,
        "com.linkedin.pegasus2avro.datajob.azkaban.AzkabanJobType": AzkabanJobTypeClass,
        "DataProcessInfo": DataProcessInfoClass,
        ".DataProcessInfo": DataProcessInfoClass,
        "com.linkedin.pegasus2avro.dataprocess.DataProcessInfo": DataProcessInfoClass,
        "DatasetDeprecation": DatasetDeprecationClass,
        ".DatasetDeprecation": DatasetDeprecationClass,
        "com.linkedin.pegasus2avro.dataset.DatasetDeprecation": DatasetDeprecationClass,
        "DatasetFieldMapping": DatasetFieldMappingClass,
        ".DatasetFieldMapping": DatasetFieldMappingClass,
        "com.linkedin.pegasus2avro.dataset.DatasetFieldMapping": DatasetFieldMappingClass,
        "DatasetLineageType": DatasetLineageTypeClass,
        ".DatasetLineageType": DatasetLineageTypeClass,
        "com.linkedin.pegasus2avro.dataset.DatasetLineageType": DatasetLineageTypeClass,
        "DatasetProperties": DatasetPropertiesClass,
        ".DatasetProperties": DatasetPropertiesClass,
        "com.linkedin.pegasus2avro.dataset.DatasetProperties": DatasetPropertiesClass,
        "DatasetUpstreamLineage": DatasetUpstreamLineageClass,
        ".DatasetUpstreamLineage": DatasetUpstreamLineageClass,
        "com.linkedin.pegasus2avro.dataset.DatasetUpstreamLineage": DatasetUpstreamLineageClass,
        "Upstream": UpstreamClass,
        ".Upstream": UpstreamClass,
        "com.linkedin.pegasus2avro.dataset.Upstream": UpstreamClass,
        "UpstreamLineage": UpstreamLineageClass,
        ".UpstreamLineage": UpstreamLineageClass,
        "com.linkedin.pegasus2avro.dataset.UpstreamLineage": UpstreamLineageClass,
        "CorpGroupInfo": CorpGroupInfoClass,
        ".CorpGroupInfo": CorpGroupInfoClass,
        "com.linkedin.pegasus2avro.identity.CorpGroupInfo": CorpGroupInfoClass,
        "CorpUserEditableInfo": CorpUserEditableInfoClass,
        ".CorpUserEditableInfo": CorpUserEditableInfoClass,
        "com.linkedin.pegasus2avro.identity.CorpUserEditableInfo": CorpUserEditableInfoClass,
        "CorpUserInfo": CorpUserInfoClass,
        ".CorpUserInfo": CorpUserInfoClass,
        "com.linkedin.pegasus2avro.identity.CorpUserInfo": CorpUserInfoClass,
        "ChartSnapshot": ChartSnapshotClass,
        ".ChartSnapshot": ChartSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.ChartSnapshot": ChartSnapshotClass,
        "CorpGroupSnapshot": CorpGroupSnapshotClass,
        ".CorpGroupSnapshot": CorpGroupSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.CorpGroupSnapshot": CorpGroupSnapshotClass,
        "CorpUserSnapshot": CorpUserSnapshotClass,
        ".CorpUserSnapshot": CorpUserSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.CorpUserSnapshot": CorpUserSnapshotClass,
        "DashboardSnapshot": DashboardSnapshotClass,
        ".DashboardSnapshot": DashboardSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.DashboardSnapshot": DashboardSnapshotClass,
        "DataFlowSnapshot": DataFlowSnapshotClass,
        ".DataFlowSnapshot": DataFlowSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.DataFlowSnapshot": DataFlowSnapshotClass,
        "DataJobSnapshot": DataJobSnapshotClass,
        ".DataJobSnapshot": DataJobSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.DataJobSnapshot": DataJobSnapshotClass,
        "DataProcessSnapshot": DataProcessSnapshotClass,
        ".DataProcessSnapshot": DataProcessSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.DataProcessSnapshot": DataProcessSnapshotClass,
        "DatasetSnapshot": DatasetSnapshotClass,
        ".DatasetSnapshot": DatasetSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.DatasetSnapshot": DatasetSnapshotClass,
        "MLFeatureSnapshot": MLFeatureSnapshotClass,
        ".MLFeatureSnapshot": MLFeatureSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.MLFeatureSnapshot": MLFeatureSnapshotClass,
        "MLModelSnapshot": MLModelSnapshotClass,
        ".MLModelSnapshot": MLModelSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.MLModelSnapshot": MLModelSnapshotClass,
        "TagSnapshot": TagSnapshotClass,
        ".TagSnapshot": TagSnapshotClass,
        "com.linkedin.pegasus2avro.metadata.snapshot.TagSnapshot": TagSnapshotClass,
        "BaseData": BaseDataClass,
        ".BaseData": BaseDataClass,
        "com.linkedin.pegasus2avro.ml.metadata.BaseData": BaseDataClass,
        "CaveatDetails": CaveatDetailsClass,
        ".CaveatDetails": CaveatDetailsClass,
        "com.linkedin.pegasus2avro.ml.metadata.CaveatDetails": CaveatDetailsClass,
        "CaveatsAndRecommendations": CaveatsAndRecommendationsClass,
        ".CaveatsAndRecommendations": CaveatsAndRecommendationsClass,
        "com.linkedin.pegasus2avro.ml.metadata.CaveatsAndRecommendations": CaveatsAndRecommendationsClass,
        "EthicalConsiderations": EthicalConsiderationsClass,
        ".EthicalConsiderations": EthicalConsiderationsClass,
        "com.linkedin.pegasus2avro.ml.metadata.EthicalConsiderations": EthicalConsiderationsClass,
        "EvaluationData": EvaluationDataClass,
        ".EvaluationData": EvaluationDataClass,
        "com.linkedin.pegasus2avro.ml.metadata.EvaluationData": EvaluationDataClass,
        "IntendedUse": IntendedUseClass,
        ".IntendedUse": IntendedUseClass,
        "com.linkedin.pegasus2avro.ml.metadata.IntendedUse": IntendedUseClass,
        "IntendedUserType": IntendedUserTypeClass,
        ".IntendedUserType": IntendedUserTypeClass,
        "com.linkedin.pegasus2avro.ml.metadata.IntendedUserType": IntendedUserTypeClass,
        "MLFeatureProperties": MLFeaturePropertiesClass,
        ".MLFeatureProperties": MLFeaturePropertiesClass,
        "com.linkedin.pegasus2avro.ml.metadata.MLFeatureProperties": MLFeaturePropertiesClass,
        "MLModelFactorPrompts": MLModelFactorPromptsClass,
        ".MLModelFactorPrompts": MLModelFactorPromptsClass,
        "com.linkedin.pegasus2avro.ml.metadata.MLModelFactorPrompts": MLModelFactorPromptsClass,
        "MLModelFactors": MLModelFactorsClass,
        ".MLModelFactors": MLModelFactorsClass,
        "com.linkedin.pegasus2avro.ml.metadata.MLModelFactors": MLModelFactorsClass,
        "MLModelProperties": MLModelPropertiesClass,
        ".MLModelProperties": MLModelPropertiesClass,
        "com.linkedin.pegasus2avro.ml.metadata.MLModelProperties": MLModelPropertiesClass,
        "Metrics": MetricsClass,
        ".Metrics": MetricsClass,
        "com.linkedin.pegasus2avro.ml.metadata.Metrics": MetricsClass,
        "QuantitativeAnalyses": QuantitativeAnalysesClass,
        ".QuantitativeAnalyses": QuantitativeAnalysesClass,
        "com.linkedin.pegasus2avro.ml.metadata.QuantitativeAnalyses": QuantitativeAnalysesClass,
        "SourceCode": SourceCodeClass,
        ".SourceCode": SourceCodeClass,
        "com.linkedin.pegasus2avro.ml.metadata.SourceCode": SourceCodeClass,
        "SourceCodeUrl": SourceCodeUrlClass,
        ".SourceCodeUrl": SourceCodeUrlClass,
        "com.linkedin.pegasus2avro.ml.metadata.SourceCodeUrl": SourceCodeUrlClass,
        "SourceCodeUrlType": SourceCodeUrlTypeClass,
        ".SourceCodeUrlType": SourceCodeUrlTypeClass,
        "com.linkedin.pegasus2avro.ml.metadata.SourceCodeUrlType": SourceCodeUrlTypeClass,
        "TrainingData": TrainingDataClass,
        ".TrainingData": TrainingDataClass,
        "com.linkedin.pegasus2avro.ml.metadata.TrainingData": TrainingDataClass,
        "MetadataChangeEvent": MetadataChangeEventClass,
        ".MetadataChangeEvent": MetadataChangeEventClass,
        "com.linkedin.pegasus2avro.mxe.MetadataChangeEvent": MetadataChangeEventClass,
        "ArrayType": ArrayTypeClass,
        ".ArrayType": ArrayTypeClass,
        "com.linkedin.pegasus2avro.schema.ArrayType": ArrayTypeClass,
        "BinaryJsonSchema": BinaryJsonSchemaClass,
        ".BinaryJsonSchema": BinaryJsonSchemaClass,
        "com.linkedin.pegasus2avro.schema.BinaryJsonSchema": BinaryJsonSchemaClass,
        "BooleanType": BooleanTypeClass,
        ".BooleanType": BooleanTypeClass,
        "com.linkedin.pegasus2avro.schema.BooleanType": BooleanTypeClass,
        "BytesType": BytesTypeClass,
        ".BytesType": BytesTypeClass,
        "com.linkedin.pegasus2avro.schema.BytesType": BytesTypeClass,
        "DatasetFieldForeignKey": DatasetFieldForeignKeyClass,
        ".DatasetFieldForeignKey": DatasetFieldForeignKeyClass,
        "com.linkedin.pegasus2avro.schema.DatasetFieldForeignKey": DatasetFieldForeignKeyClass,
        "DateType": DateTypeClass,
        ".DateType": DateTypeClass,
        "com.linkedin.pegasus2avro.schema.DateType": DateTypeClass,
        "EditableSchemaFieldInfo": EditableSchemaFieldInfoClass,
        ".EditableSchemaFieldInfo": EditableSchemaFieldInfoClass,
        "com.linkedin.pegasus2avro.schema.EditableSchemaFieldInfo": EditableSchemaFieldInfoClass,
        "EditableSchemaMetadata": EditableSchemaMetadataClass,
        ".EditableSchemaMetadata": EditableSchemaMetadataClass,
        "com.linkedin.pegasus2avro.schema.EditableSchemaMetadata": EditableSchemaMetadataClass,
        "EnumType": EnumTypeClass,
        ".EnumType": EnumTypeClass,
        "com.linkedin.pegasus2avro.schema.EnumType": EnumTypeClass,
        "EspressoSchema": EspressoSchemaClass,
        ".EspressoSchema": EspressoSchemaClass,
        "com.linkedin.pegasus2avro.schema.EspressoSchema": EspressoSchemaClass,
        "FixedType": FixedTypeClass,
        ".FixedType": FixedTypeClass,
        "com.linkedin.pegasus2avro.schema.FixedType": FixedTypeClass,
        "ForeignKeySpec": ForeignKeySpecClass,
        ".ForeignKeySpec": ForeignKeySpecClass,
        "com.linkedin.pegasus2avro.schema.ForeignKeySpec": ForeignKeySpecClass,
        "KafkaSchema": KafkaSchemaClass,
        ".KafkaSchema": KafkaSchemaClass,
        "com.linkedin.pegasus2avro.schema.KafkaSchema": KafkaSchemaClass,
        "KeyValueSchema": KeyValueSchemaClass,
        ".KeyValueSchema": KeyValueSchemaClass,
        "com.linkedin.pegasus2avro.schema.KeyValueSchema": KeyValueSchemaClass,
        "MapType": MapTypeClass,
        ".MapType": MapTypeClass,
        "com.linkedin.pegasus2avro.schema.MapType": MapTypeClass,
        "MySqlDDL": MySqlDDLClass,
        ".MySqlDDL": MySqlDDLClass,
        "com.linkedin.pegasus2avro.schema.MySqlDDL": MySqlDDLClass,
        "NullType": NullTypeClass,
        ".NullType": NullTypeClass,
        "com.linkedin.pegasus2avro.schema.NullType": NullTypeClass,
        "NumberType": NumberTypeClass,
        ".NumberType": NumberTypeClass,
        "com.linkedin.pegasus2avro.schema.NumberType": NumberTypeClass,
        "OracleDDL": OracleDDLClass,
        ".OracleDDL": OracleDDLClass,
        "com.linkedin.pegasus2avro.schema.OracleDDL": OracleDDLClass,
        "OrcSchema": OrcSchemaClass,
        ".OrcSchema": OrcSchemaClass,
        "com.linkedin.pegasus2avro.schema.OrcSchema": OrcSchemaClass,
        "OtherSchema": OtherSchemaClass,
        ".OtherSchema": OtherSchemaClass,
        "com.linkedin.pegasus2avro.schema.OtherSchema": OtherSchemaClass,
        "PrestoDDL": PrestoDDLClass,
        ".PrestoDDL": PrestoDDLClass,
        "com.linkedin.pegasus2avro.schema.PrestoDDL": PrestoDDLClass,
        "RecordType": RecordTypeClass,
        ".RecordType": RecordTypeClass,
        "com.linkedin.pegasus2avro.schema.RecordType": RecordTypeClass,
        "SchemaField": SchemaFieldClass,
        ".SchemaField": SchemaFieldClass,
        "com.linkedin.pegasus2avro.schema.SchemaField": SchemaFieldClass,
        "SchemaFieldDataType": SchemaFieldDataTypeClass,
        ".SchemaFieldDataType": SchemaFieldDataTypeClass,
        "com.linkedin.pegasus2avro.schema.SchemaFieldDataType": SchemaFieldDataTypeClass,
        "SchemaMetadata": SchemaMetadataClass,
        ".SchemaMetadata": SchemaMetadataClass,
        "com.linkedin.pegasus2avro.schema.SchemaMetadata": SchemaMetadataClass,
        "Schemaless": SchemalessClass,
        ".Schemaless": SchemalessClass,
        "com.linkedin.pegasus2avro.schema.Schemaless": SchemalessClass,
        "StringType": StringTypeClass,
        ".StringType": StringTypeClass,
        "com.linkedin.pegasus2avro.schema.StringType": StringTypeClass,
        "TimeType": TimeTypeClass,
        ".TimeType": TimeTypeClass,
        "com.linkedin.pegasus2avro.schema.TimeType": TimeTypeClass,
        "UnionType": UnionTypeClass,
        ".UnionType": UnionTypeClass,
        "com.linkedin.pegasus2avro.schema.UnionType": UnionTypeClass,
        "UrnForeignKey": UrnForeignKeyClass,
        ".UrnForeignKey": UrnForeignKeyClass,
        "com.linkedin.pegasus2avro.schema.UrnForeignKey": UrnForeignKeyClass,
        "TagProperties": TagPropertiesClass,
        ".TagProperties": TagPropertiesClass,
        "com.linkedin.pegasus2avro.tag.TagProperties": TagPropertiesClass,
    }
    
    
    def __init__(self, readers_schema=None, **kwargs):
        writers_schema = kwargs.pop("writers_schema", readers_schema)
        writers_schema = kwargs.pop("writer_schema", writers_schema)
        super(SpecificDatumReader, self).__init__(writers_schema, readers_schema, **kwargs)
    
    
    def read_record(self, writers_schema, readers_schema, decoder):
        result = super(SpecificDatumReader, self).read_record(writers_schema, readers_schema, decoder)
        
        if readers_schema.fullname in SpecificDatumReader.SCHEMA_TYPES:
            result = SpecificDatumReader.SCHEMA_TYPES[readers_schema.fullname](result)
        
        return result# fmt: on



import json
from typing import TYPE_CHECKING, Dict, List, Optional

import dateutil.parser
from airflow.lineage.backend import LineageBackend

import datahub.emitter.mce_builder as builder
import datahub.metadata as models

if TYPE_CHECKING:
    from airflow import DAG

    # from airflow.taskinstance import TaskInstance
    from airflow.models.baseoperator import (
        BaseOperator,
    )  # pylint: disable=cyclic-import

from airflow.configuration import conf


def _entities_to_urn_list(iolets: List):
    return [let.urn for let in iolets]


def make_emitter_hook():
    # This is necessary to avoid issues with circular imports.
    from datahub.integrations.airflow.hooks import DatahubGenericHook

    _datahub_conn_id = conf.get("lineage", "datahub_conn_id")
    return DatahubGenericHook(_datahub_conn_id)


class DatahubAirflowLineageBackend(LineageBackend):
    # With Airflow 2.0, this can be an instance method. However, with Airflow 1.10.x, this
    # method is used statically, even though LineageBackend declares it as an instance variable.
    @staticmethod
    def send_lineage(
        operator: "BaseOperator",
        inlets: Optional[List] = None,
        outlets: Optional[List] = None,
        context: Dict = None,
    ):
        context = context or {}  # ensure not None to satisfy mypy

        dag: "DAG" = context["dag"]
        task = context["task"]
        # task_instance: "TaskInstance" = context["task_instance"]

        # TODO: verify if task and operator are the same?
        # TODO: use dag serialization to just save the whole thing.
        # TODO: save context.get("conf")
        # TODO: save DAG tags
        # TODO: save context.get("dag_run")
        # TODO: save all the data from task_instance
        # TODO: capture raw sql from db operators

        flow_urn = builder.make_data_flow_urn("airflow", dag.dag_id)
        job_urn = builder.make_data_job_urn_with_flow(flow_urn, task.task_id)

        timestamp = int(dateutil.parser.parse(context["ts"]).timestamp() * 1000)
        ownership = models.OwnershipClass(
            owners=[
                models.OwnerClass(
                    owner=dag.owner,
                    type=models.OwnershipTypeClass.DEVELOPER,
                    source=models.OwnershipSourceClass(
                        type=models.OwnershipSourceTypeClass.SERVICE,
                        url=dag.filepath,
                    ),
                )
            ],
            lastModified=models.AuditStampClass(
                time=timestamp, actor=builder.make_user_urn("airflow")
            ),
        )

        flow_mce = models.MetadataChangeEventClass(
            proposedSnapshot=models.DataFlowSnapshotClass(
                urn=flow_urn,
                aspects=[
                    models.DataFlowInfoClass(
                        name=dag.dag_id,
                        description=f"{dag.description}\n\n{dag.doc_md}",
                    ),
                    ownership,
                ],
            )
        )

        job_mce = models.MetadataChangeEventClass(
            proposedSnapshot=models.DataJobSnapshotClass(
                urn=job_urn,
                aspects=[
                    models.DataJobInfoClass(
                        name=task.task_id,
                        type=models.AzkabanJobTypeClass.COMMAND,
                        description=None,  # TODO: add datajob description
                    ),
                    models.DataJobInputOutputClass(
                        inputDatasets=_entities_to_urn_list(inlets or []),
                        outputDatasets=_entities_to_urn_list(outlets or []),
                    ),
                    ownership,
                ],
            )
        )

        lineage_mces = [
            builder.make_lineage_mce(_entities_to_urn_list(inlets or []), outlet)
            for outlet in _entities_to_urn_list(outlets or [])
        ]

        hook = make_emitter_hook()

        mces = [
            flow_mce,
            job_mce,
            *lineage_mces,
        ]
        operator.log.info(
            "DataHub lineage backend - emitting metadata:\n"
            + "\n".join(json.dumps(mce.to_obj()) for mce in mces)
        )
        hook.emit_mces(mces)



from collections import OrderedDict
from typing import Any, Dict, Type

import requests
from requests.exceptions import HTTPError, RequestException

from datahub.configuration.common import OperationalError
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.schema_classes import (  # MLFeatureSnapshotClass,
    ChartSnapshotClass,
    CorpGroupSnapshotClass,
    CorpUserSnapshotClass,
    DashboardSnapshotClass,
    DataFlowSnapshotClass,
    DataJobSnapshotClass,
    DataProcessSnapshotClass,
    DatasetSnapshotClass,
    MLModelSnapshotClass,
    TagSnapshotClass,
)

resource_locator: Dict[Type[object], str] = {
    ChartSnapshotClass: "charts",
    DashboardSnapshotClass: "dashboards",
    CorpUserSnapshotClass: "corpUsers",
    CorpGroupSnapshotClass: "corpGroups",
    DatasetSnapshotClass: "datasets",
    DataProcessSnapshotClass: "dataProcesses",
    DataFlowSnapshotClass: "dataFlows",
    DataJobSnapshotClass: "dataJobs",
    MLModelSnapshotClass: "mlModels",
    TagSnapshotClass: "tags",
}


def _rest_li_ify(obj: Any) -> Any:
    if isinstance(obj, (dict, OrderedDict)):
        if len(obj.keys()) == 1:
            key = list(obj.keys())[0]
            value = obj[key]
            if key.find("com.linkedin.pegasus2avro.") >= 0:
                new_key = key.replace("com.linkedin.pegasus2avro.", "com.linkedin.")
                return {new_key: _rest_li_ify(value)}

        new_obj: Any = {}
        for key, value in obj.items():
            if value is not None:
                new_obj[key] = _rest_li_ify(value)
        return new_obj
    elif isinstance(obj, list):
        new_obj = [_rest_li_ify(item) for item in obj]
        return new_obj
    return obj


class DatahubRestEmitter:
    _gms_server: str

    def __init__(self, gms_server: str):
        self._gms_server = gms_server

    def _get_ingest_endpoint(self, mce: MetadataChangeEvent) -> str:
        snapshot_type = type(mce.proposedSnapshot)
        snapshot_resource = resource_locator.get(snapshot_type, None)
        if not snapshot_resource:
            raise ValueError(
                f"Failed to locate a snapshot resource for type {snapshot_type}"
            )

        return f"{self._gms_server}/{snapshot_resource}?action=ingest"

    def emit_mce(self, mce: MetadataChangeEvent) -> None:
        url = self._get_ingest_endpoint(mce)
        headers = {"X-RestLi-Protocol-Version": "2.0.0"}

        raw_mce_obj = mce.proposedSnapshot.to_obj()
        mce_obj = _rest_li_ify(raw_mce_obj)
        snapshot = {"snapshot": mce_obj}

        try:
            response = requests.post(url, headers=headers, json=snapshot)

            # import curlify
            # print(curlify.to_curl(response.request))
            # breakpoint()

            response.raise_for_status()
        except HTTPError as e:
            info = response.json()
            raise OperationalError(
                "Unable to emit metadata to DataHub GMS", info
            ) from e
        except RequestException as e:
            raise OperationalError(
                "Unable to emit metadata to DataHub GMS", {"message": str(e)}
            ) from e



import time
from dataclasses import dataclass
from dataclasses import field as dataclass_field
from typing import Dict, Iterable, List, Optional

import boto3

from datahub.configuration import ConfigModel
from datahub.configuration.common import AllowDenyPattern
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.source.metadata_common import MetadataWorkUnit
from datahub.metadata.com.linkedin.pegasus2avro.common import AuditStamp, Status
from datahub.metadata.com.linkedin.pegasus2avro.metadata.snapshot import DatasetSnapshot
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    ArrayTypeClass,
    BooleanTypeClass,
    BytesTypeClass,
    DateTypeClass,
    MySqlDDL,
    NullTypeClass,
    NumberTypeClass,
    SchemaField,
    SchemaFieldDataType,
    SchemaMetadata,
    StringTypeClass,
    TimeTypeClass,
    UnionTypeClass,
)
from datahub.metadata.schema_classes import (
    AuditStampClass,
    DatasetPropertiesClass,
    MapTypeClass,
    OwnerClass,
    OwnershipClass,
    OwnershipTypeClass,
)


class GlueSourceConfig(ConfigModel):
    env: str = "PROD"
    database_pattern: AllowDenyPattern = AllowDenyPattern.allow_all()
    table_pattern: AllowDenyPattern = AllowDenyPattern.allow_all()
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_session_token: Optional[str] = None
    aws_region: str

    @property
    def glue_client(self):
        if (
            self.aws_access_key_id
            and self.aws_secret_access_key
            and self.aws_session_token
        ):
            return boto3.client(
                "glue",
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
                region_name=self.aws_region,
            )
        elif self.aws_access_key_id and self.aws_secret_access_key:
            return boto3.client(
                "glue",
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.aws_region,
            )
        else:
            return boto3.client("glue", region_name=self.aws_region)


@dataclass
class GlueSourceReport(SourceReport):
    tables_scanned = 0
    filtered: List[str] = dataclass_field(default_factory=list)

    def report_table_scanned(self) -> None:
        self.tables_scanned += 1

    def report_table_dropped(self, table: str) -> None:
        self.filtered.append(table)


class GlueSource(Source):
    source_config: GlueSourceConfig
    report = GlueSourceReport()

    def __init__(self, config: GlueSourceConfig, ctx: PipelineContext):
        super().__init__(ctx)
        self.source_config = config
        self.report = GlueSourceReport()
        self.glue_client = config.glue_client
        self.env = config.env

    @classmethod
    def create(cls, config_dict, ctx):
        config = GlueSourceConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def get_workunits(self) -> Iterable[MetadataWorkUnit]:
        def get_all_tables():
            def get_tables_from_database(database_name: str, tables: List):
                kwargs = {"DatabaseName": database_name}
                while True:
                    data = self.glue_client.get_tables(**kwargs)
                    tables += data["TableList"]
                    if "NextToken" in data:
                        kwargs["NextToken"] = data["NextToken"]
                    else:
                        break
                return tables

            def get_tables_from_all_databases():
                tables = []
                kwargs: Dict = {}
                while True:
                    data = self.glue_client.search_tables(**kwargs)
                    tables += data["TableList"]
                    if "NextToken" in data:
                        kwargs["NextToken"] = data["NextToken"]
                    else:
                        break
                return tables

            if self.source_config.database_pattern.is_fully_specified_allow_list():
                all_tables: List = []
                database_names = self.source_config.database_pattern.get_allowed_list()
                for database in database_names:
                    all_tables += get_tables_from_database(database, all_tables)
            else:
                all_tables = get_tables_from_all_databases()
            return all_tables

        tables = get_all_tables()

        for table in tables:
            database_name = table["DatabaseName"]
            table_name = table["Name"]
            full_table_name = f"{database_name}.{table_name}"
            self.report.report_table_scanned()
            if not self.source_config.database_pattern.allowed(
                database_name
            ) or not self.source_config.table_pattern.allowed(full_table_name):
                self.report.report_table_dropped(full_table_name)
                continue

            mce = self._extract_record(table, full_table_name)
            workunit = MetadataWorkUnit(id=f"glue-{full_table_name}", mce=mce)
            self.report.report_workunit(workunit)
            yield workunit

    def _extract_record(self, table: Dict, table_name: str) -> MetadataChangeEvent:
        def get_owner(time) -> OwnershipClass:
            owner = table.get("Owner")
            if owner:
                owners = [
                    OwnerClass(
                        owner=f"urn:li:corpuser:{owner}",
                        type=OwnershipTypeClass.DATAOWNER,
                    )
                ]
            else:
                owners = []
            return OwnershipClass(
                owners=owners,
                lastModified=AuditStampClass(
                    time=time,
                    actor="urn:li:corpuser:datahub",
                ),
            )

        def get_dataset_properties() -> DatasetPropertiesClass:
            return DatasetPropertiesClass(
                description=table.get("Description"),
                customProperties={
                    **table.get("Parameters", {}),
                    **{
                        k: str(v)
                        for k, v in table["StorageDescriptor"].items()
                        if k not in ["Columns", "Parameters"]
                    },
                },
                uri=table.get("Location"),
                tags=[],
            )

        def get_schema_metadata(glue_source: GlueSource):
            schema = table["StorageDescriptor"]["Columns"]
            fields: List[SchemaField] = []
            for field in schema:
                schema_field = SchemaField(
                    fieldPath=field["Name"],
                    nativeDataType=field["Type"],
                    type=get_column_type(
                        glue_source, field["Type"], table_name, field["Name"]
                    ),
                    description=field.get("Comment"),
                    recursive=False,
                    nullable=True,
                )
                fields.append(schema_field)
            return SchemaMetadata(
                schemaName=table_name,
                version=0,
                fields=fields,
                platform="urn:li:dataPlatform:glue",
                created=AuditStamp(time=sys_time, actor="urn:li:corpuser:etl"),
                lastModified=AuditStamp(time=sys_time, actor="urn:li:corpuser:etl"),
                hash="",
                platformSchema=MySqlDDL(tableSchema=""),
            )

        sys_time = int(time.time() * 1000)
        metadata_record = MetadataChangeEvent()
        dataset_snapshot = DatasetSnapshot(
            urn=f"urn:li:dataset:(urn:li:dataPlatform:glue,{table_name},{self.env})",
            aspects=[],
        )

        dataset_snapshot.aspects.append(Status(removed=False))
        dataset_snapshot.aspects.append(get_owner(sys_time))
        dataset_snapshot.aspects.append(get_dataset_properties())
        dataset_snapshot.aspects.append(get_schema_metadata(self))

        metadata_record.proposedSnapshot = dataset_snapshot

        return metadata_record

    def get_report(self):
        return self.report

    def close(self):
        pass


def get_column_type(
    glue_source: GlueSource, field_type: str, table_name: str, field_name: str
) -> SchemaFieldDataType:
    field_type_mapping = {
        "array": ArrayTypeClass,
        "bigint": NumberTypeClass,
        "binary": BytesTypeClass,
        "boolean": BooleanTypeClass,
        "char": StringTypeClass,
        "date": DateTypeClass,
        "decimal": NumberTypeClass,
        "double": NumberTypeClass,
        "float": NumberTypeClass,
        "int": NumberTypeClass,
        "integer": NumberTypeClass,
        "interval": TimeTypeClass,
        "long": NumberTypeClass,
        "map": MapTypeClass,
        "null": NullTypeClass,
        "set": ArrayTypeClass,
        "smallint": NumberTypeClass,
        "string": StringTypeClass,
        "struct": MapTypeClass,
        "timestamp": TimeTypeClass,
        "tinyint": NumberTypeClass,
        "union": UnionTypeClass,
        "varchar": StringTypeClass,
    }

    if field_type in field_type_mapping.keys():
        type_class = field_type_mapping[field_type]
    elif field_type.startswith("array"):
        type_class = ArrayTypeClass
    elif field_type.startswith("map") or field_type.startswith("struct"):
        type_class = MapTypeClass
    elif field_type.startswith("set"):
        type_class = ArrayTypeClass
    else:
        glue_source.report.report_warning(
            field_type,
            f"The type '{field_type}' is not recognised for field '{field_name}' in table '{table_name}', setting as StringTypeClass.",
        )
        type_class = StringTypeClass
    data_type = SchemaFieldDataType(type=type_class())
    return data_type



from typing import List, Union

from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

from datahub.integrations.airflow.hooks import (
    DatahubGenericHook,
    DatahubKafkaHook,
    DatahubRestHook,
)
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent


class DatahubBaseOperator(BaseOperator):
    ui_color = "#4398c8"

    hook: Union[DatahubRestHook, DatahubKafkaHook]

    @apply_defaults
    def __init__(
        self,
        *,
        datahub_conn_id: str,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.datahub_conn_id = datahub_conn_id
        self.generic_hook = DatahubGenericHook(datahub_conn_id)


class DatahubEmitterOperator(DatahubBaseOperator):
    @apply_defaults
    def __init__(
        self,
        mces: List[MetadataChangeEvent],
        datahub_conn_id: str,
        **kwargs,
    ):
        super().__init__(
            datahub_conn_id=datahub_conn_id,
            **kwargs,
        )
        self.mces = mces

    def execute(self, context):
        self.generic_hook.get_underlying_hook().emit_mces(self.mces)



from airflow.plugins_manager import AirflowPlugin

import datahub
from datahub.integrations.airflow.hooks import DatahubKafkaHook, DatahubRestHook


def get_provider_info():
    return {
        "name": "DataHub",
        "description": "`DataHub <https://datahubproject.io/>`__\n",
        "hook-class-names": [
            "datahub.integrations.airflow.hooks.DatahubRestHook",
            "datahub.integrations.airflow.hooks.DatahubKafkaHook",
        ],
        "package-name": datahub.__package_name__,
        "versions": [datahub.__version__],
    }


class DatahubAirflowPlugin(AirflowPlugin):
    # TODO: determine if this is actually necessary.
    # We only want to register the hooks here, so that we can be sure they show up
    # in the Web UI. Registering operators via plugins is deprecated.
    name = "datahub"
    hooks = [DatahubKafkaHook, DatahubRestHook]



import logging
from dataclasses import dataclass

from datahub.configuration.common import ConfigModel, OperationalError
from datahub.emitter.rest_emitter import DatahubRestEmitter
from datahub.ingestion.api.common import PipelineContext, RecordEnvelope, WorkUnit
from datahub.ingestion.api.sink import Sink, SinkReport, WriteCallback
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent

logger = logging.getLogger(__name__)


class DatahubRestSinkConfig(ConfigModel):
    """Configuration class for holding connectivity to datahub gms"""

    server: str = "http://localhost:8080"


@dataclass
class DatahubRestSink(Sink):
    config: DatahubRestSinkConfig
    emitter: DatahubRestEmitter
    report: SinkReport

    def __init__(self, ctx: PipelineContext, config: DatahubRestSinkConfig):
        super().__init__(ctx)
        self.config = config
        self.report = SinkReport()
        self.emitter = DatahubRestEmitter(self.config.server)

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext):
        config = DatahubRestSinkConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def handle_work_unit_start(self, workunit: WorkUnit) -> None:
        pass

    def handle_work_unit_end(self, workunit: WorkUnit) -> None:
        pass

    def write_record_async(
        self,
        record_envelope: RecordEnvelope[MetadataChangeEvent],
        write_callback: WriteCallback,
    ):
        mce = record_envelope.record

        try:
            self.emitter.emit_mce(mce)
            self.report.report_record_written(record_envelope)
            write_callback.on_success(record_envelope, {})
        except OperationalError as e:
            self.report.report_failure({"error": e.message, "info": e.info})
            write_callback.on_failure(record_envelope, e, e.info)
        except Exception as e:
            self.report.report_failure({"e": e})
            write_callback.on_failure(record_envelope, e, {})

    def get_report(self) -> SinkReport:
        return self.report

    def close(self):
        pass



from dataclasses import dataclass, field
from typing import Iterable, List, Optional

import pymongo

from datahub.configuration.common import AllowDenyPattern, ConfigModel
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.source.metadata_common import MetadataWorkUnit
from datahub.metadata.com.linkedin.pegasus2avro.metadata.snapshot import DatasetSnapshot
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.schema_classes import DatasetPropertiesClass

# These are MongoDB-internal databases, which we want to skip.
# See https://docs.mongodb.com/manual/reference/local-database/ and
# https://docs.mongodb.com/manual/reference/config-database/ and
# https://stackoverflow.com/a/48273736/5004662.
DENY_DATABASE_LIST = set(["admin", "config", "local"])


class MongoDBConfig(ConfigModel):
    # See the MongoDB authentication docs for details and examples.
    # https://pymongo.readthedocs.io/en/stable/examples/authentication.html
    connect_uri: str = "mongodb://localhost"
    username: Optional[str] = None
    password: Optional[str] = None
    authMechanism: Optional[str] = None
    options: dict = {}

    database_pattern: AllowDenyPattern = AllowDenyPattern.allow_all()
    collection_pattern: AllowDenyPattern = AllowDenyPattern.allow_all()


@dataclass
class MongoDBSourceReport(SourceReport):
    filtered: List[str] = field(default_factory=list)

    def report_dropped(self, name: str) -> None:
        self.filtered.append(name)


@dataclass
class MongoDBSource(Source):
    config: MongoDBConfig
    report: MongoDBSourceReport

    def __init__(self, ctx: PipelineContext, config: MongoDBConfig):
        super().__init__(ctx)
        self.config = config
        self.report = MongoDBSourceReport()

        options = {}
        if self.config.username is not None:
            options["username"] = self.config.username
        if self.config.password is not None:
            options["password"] = self.config.password
        if self.config.authMechanism is not None:
            options["authMechanism"] = self.config.authMechanism
        options = {
            **options,
            **self.config.options,
        }

        self.mongo_client = pymongo.MongoClient(self.config.connect_uri, **options)

        # This cheaply tests the connection. For details, see
        # https://pymongo.readthedocs.io/en/stable/api/pymongo/mongo_client.html#pymongo.mongo_client.MongoClient
        self.mongo_client.admin.command("ismaster")

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext):
        config = MongoDBConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_workunits(self) -> Iterable[MetadataWorkUnit]:
        env = "PROD"
        platform = "mongodb"

        database_names: List[str] = self.mongo_client.list_database_names()
        for database_name in database_names:
            if database_name in DENY_DATABASE_LIST:
                continue
            if not self.config.database_pattern.allowed(database_name):
                self.report.report_dropped(database_name)
                continue

            database = self.mongo_client[database_name]
            collection_names: List[str] = database.list_collection_names()
            for collection_name in collection_names:
                dataset_name = f"{database_name}.{collection_name}"
                if not self.config.collection_pattern.allowed(dataset_name):
                    self.report.report_dropped(dataset_name)
                    continue

                mce = MetadataChangeEvent()
                dataset_snapshot = DatasetSnapshot()
                dataset_snapshot.urn = f"urn:li:dataset:(urn:li:dataPlatform:{platform},{dataset_name},{env})"

                dataset_properties = DatasetPropertiesClass(
                    tags=[],
                    customProperties={},
                )
                dataset_snapshot.aspects.append(dataset_properties)

                # TODO: Guess the schema via sampling
                # State of the art seems to be https://github.com/variety/variety.

                # TODO: use list_indexes() or index_information() to get index information
                # See https://pymongo.readthedocs.io/en/stable/api/pymongo/collection.html#pymongo.collection.Collection.list_indexes.

                mce.proposedSnapshot = dataset_snapshot

                wu = MetadataWorkUnit(id=dataset_name, mce=mce)
                self.report.report_workunit(wu)
                yield wu

    def get_report(self) -> MongoDBSourceReport:
        return self.report

    def close(self):
        self.mongo_client.close()



import os
from typing import Dict, Set

import setuptools

package_metadata: dict = {}
with open("./src/datahub/__init__.py") as fp:
    exec(fp.read(), package_metadata)


def get_long_description():
    root = os.path.dirname(__file__)
    with open(os.path.join(root, "README.md")) as f:
        description = f.read()

    return description


base_requirements = {
    # Compatability.
    "dataclasses>=0.6; python_version < '3.7'",
    "typing_extensions>=3.7.4; python_version < '3.8'",
    "mypy_extensions>=0.4.3",
    # Actual dependencies.
    "typing-inspect",
    "pydantic>=1.5.1",
}

framework_common = {
    "click>=6.0.0",
    "PyYAML",
    "toml>=0.10.0",
    "docker",
    "expandvars>=0.6.5",
    "avro-gen3==0.3.9",
    "avro-python3>=1.8.2",
    "python-dateutil",
}

kafka_common = {
    # We currently require both Avro libraries. The codegen uses avro-python3 (above)
    # schema parsers at runtime for generating and reading JSON into Python objects.
    # At the same time, we use Kafka's AvroSerializer, which internally relies on
    # fastavro for serialization. We do not use confluent_kafka[avro], since it
    # is incompatible with its own dep on avro-python3.
    "confluent_kafka>=1.5.0",
    "fastavro>=1.2.0",
}

sql_common = {
    # Required for all SQL sources.
    "sqlalchemy>=1.3.24",
}

# Note: for all of these, framework_common will be added.
plugins: Dict[str, Set[str]] = {
    # Sink plugins.
    "datahub-kafka": kafka_common,
    "datahub-rest": {"requests"},
    # Integrations.
    "airflow": {"apache-airflow >= 1.10.2"},
    # Source plugins
    "kafka": kafka_common,
    "athena": sql_common | {"PyAthena[SQLAlchemy]"},
    "bigquery": sql_common | {"pybigquery >= 0.6.0"},
    "hive": sql_common | {"pyhive[hive]"},
    "mssql": sql_common | {"sqlalchemy-pytds>=0.3"},
    "mysql": sql_common | {"pymysql>=1.0.2"},
    "postgres": sql_common | {"psycopg2-binary", "GeoAlchemy2"},
    "snowflake": sql_common | {"snowflake-sqlalchemy"},
    "oracle": sql_common | {"cx_Oracle"},
    "ldap": {"python-ldap>=2.4"},
    "druid": sql_common | {"pydruid>=0.6.2"},
    "mongodb": {"pymongo>=3.11"},
    "glue": {"boto3"},
}

dev_requirements = {
    *base_requirements,
    *framework_common,
    "black>=19.10b0",
    "coverage>=5.1",
    "flake8>=3.8.3",
    "isort>=5.7.0",
    "mypy>=0.782",
    "pytest>=6.2.2",
    "pytest-cov>=2.8.1",
    "pytest-docker",
    "sqlalchemy-stubs",
    "deepdiff",
    "freezegun",
    "build",
    "twine",
    # Also add the plugins which are used for tests.
    "apache-airflow==1.10.15",  # Airflow 2.x does not have LineageBackend packaged yet.
    "apache-airflow-backport-providers-snowflake",  # Used in the example DAGs.
    *list(
        dependency
        for plugin in [
            "bigquery",
            "mysql",
            "mssql",
            "mongodb",
            "ldap",
            "glue",
            "datahub-kafka",
            "datahub-rest",
            "airflow",
        ]
        for dependency in plugins[plugin]
    ),
}


setuptools.setup(
    # Package metadata.
    name=package_metadata["__package_name__"],
    version=package_metadata["__version__"],
    url="https://datahubproject.io/",
    project_urls={
        "Documentation": "https://datahubproject.io/docs/",
        "Source": "https://github.com/linkedin/datahub",
        "Changelog": "https://github.com/linkedin/datahub/releases",
    },
    author="DataHub Committers",
    license="Apache License 2.0",
    description="A CLI to work with DataHub metadata",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Unix",
        "Operating System :: POSIX :: Linux",
        "Environment :: Console",
        "Environment :: MacOS X",
        "Topic :: Software Development",
    ],
    # Package info.
    zip_safe=False,
    python_requires=">=3.6",
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="./src"),
    include_package_data=True,
    package_data={
        "datahub": ["py.typed"],
        "datahub.metadata": ["schema.avsc"],
    },
    entry_points={
        "console_scripts": ["datahub = datahub.entrypoints:datahub"],
        "datahub.ingestion.source.plugins": [
            "file = datahub.ingestion.source.mce_file:MetadataFileSource",
            "sqlalchemy = datahub.ingestion.source.sql_generic:SQLAlchemyGenericSource",
            "athena = datahub.ingestion.source.athena:AthenaSource",
            "bigquery = datahub.ingestion.source.bigquery:BigQuerySource",
            "dbt = datahub.ingestion.source.dbt:DBTSource",
            "druid = datahub.ingestion.source.druid:DruidSource",
            "glue = datahub.ingestion.source.glue:GlueSource",
            "hive = datahub.ingestion.source.hive:HiveSource",
            "kafka = datahub.ingestion.source.kafka:KafkaSource",
            "ldap = datahub.ingestion.source.ldap:LDAPSource",
            "mongodb = datahub.ingestion.source.mongodb:MongoDBSource",
            "mssql = datahub.ingestion.source.mssql:SQLServerSource",
            "mysql = datahub.ingestion.source.mysql:MySQLSource",
            "oracle = datahub.ingestion.source.oracle:OracleSource",
            "postgres = datahub.ingestion.source.postgres:PostgresSource",
            "snowflake = datahub.ingestion.source.snowflake:SnowflakeSource",
        ],
        "datahub.ingestion.sink.plugins": [
            "file = datahub.ingestion.sink.file:FileSink",
            "console = datahub.ingestion.sink.console:ConsoleSink",
            "datahub-kafka = datahub.ingestion.sink.datahub_kafka:DatahubKafkaSink",
            "datahub-rest = datahub.ingestion.sink.datahub_rest:DatahubRestSink",
        ],
        "apache_airflow_provider": [
            "provider_info=datahub.integrations.airflow.get_provider_info:get_provider_info"
        ],
        "airflow.plugins": [
            "datahub = datahub.integrations.airflow.get_provider_info:DatahubAirflowPlugin"
        ],
    },
    # Dependencies.
    install_requires=list(base_requirements | framework_common),
    extras_require={
        "base": list(framework_common),
        **{
            plugin: list(framework_common | dependencies)
            for (plugin, dependencies) in plugins.items()
        },
        "all": list(framework_common.union(*plugins.values())),
        "dev": list(dev_requirements),
    },
)



from typing import IO

import toml

from .common import ConfigurationMechanism


class TomlConfigurationMechanism(ConfigurationMechanism):
    """Ability to load configuration from toml files"""

    def load_config(self, config_fp: IO):
        config = toml.load(config_fp)
        return config



__package_name__ = "acryl-datahub"
__version__ = "0.1.1"


def nice_version_name() -> str:
    if __version__ == "0.0.0.dev0":
        return "unavailable (installed editable via git)"
    return __version__
